#include <string.h>
#include <sys/time.h>

static char *fmt = "%l:%M%P on %A, %d %B %Y";

int
get_date(cp)
	char *cp;
{
	time_t tm;

	time (&tm);
	strftime (cp, 50, fmt, localtime (&tm));
	return 0;
}
